#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
""" Functions to convert QuantizedStatefulRecurrent to Akida. """

import quantizeml.layers as qlayers
import numpy as np

from akida import StatefulRecurrent, ActivationType

from ..akida_versions import AkidaVersion
from .activations import set_relu_variables, parse_relu_v2
from .block_converter import BlockConverter, register_conversion_patterns
from .blocks import get_block_out_quantizer
from .conv_common import get_layer_by_type
from .outputs import set_output_v2_variables, parse_output_bits, parse_post_op_buffer_bits
from .weights import broadcast_and_set_variable

__all__ = ["StatefulRecurrentBlockConverter"]

_PATTERNS = [(qlayers.QuantizedStatefulProjection, qlayers.QuantizedStatefulRecurrent,
              qlayers.QuantizedExtractToken, qlayers.QuantizedStatefulProjection,
              qlayers.QuantizedReLU),
             (qlayers.QuantizedStatefulProjection, qlayers.QuantizedStatefulRecurrent,
              qlayers.QuantizedExtractToken, qlayers.QuantizedStatefulProjection)]


def _set_stateful_recurrent_variables(ak_layer, block):
    """Computes and sets the variables for an Akida StatefulRecurrent layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        block (list(:obj:`keras.Layer`)): the block of keras layers.
    """
    variables_ak = ak_layer.variables

    input_proj = block[0]
    stateful_rec = block[1]
    output_proj = block[3]

    assert isinstance(input_proj.weight_quantizer, qlayers.WeightQuantizer)
    assert isinstance(output_proj.weight_quantizer, qlayers.WeightQuantizer)
    assert isinstance(stateful_rec.a_quantizer, qlayers.WeightQuantizer)

    # Set StatefulRecurrent input_shift
    input_shift = getattr(input_proj, 'input_shift', None)
    if input_shift is not None:
        broadcast_and_set_variable(variables_ak, "input_shift",
                                   input_shift.value.numpy().astype(np.uint8))
    # Set input_projection dense kernel
    in_proj_ak = input_proj.weight_quantizer.qweights.value.fp.values.numpy()
    variables_ak["in_proj"] = in_proj_ak.astype(np.int8)

    assert input_proj.out_quantizer is None, "Unexpected out_quantizer on input projection."

    # Set new_state_shift, new_state_scales that is actually recorded in stateful_rec
    if new_state_shift := getattr(stateful_rec, 'new_state_shift', None):
        broadcast_and_set_variable(variables_ak, "new_state_shift",
                                   new_state_shift.value.numpy().astype(np.int8))

    if new_state_scale := getattr(stateful_rec, 'new_state_scale', None):
        broadcast_and_set_variable(variables_ak, "new_state_scale",
                                   new_state_scale.value.values.numpy().astype(np.uint8))

    # Set A weights variables
    A_real_ak = stateful_rec.a_quantizer(stateful_rec.A_real).values.numpy()
    variables_ak["A_real"] = A_real_ak.astype(np.int16)
    A_imag_ak = stateful_rec.a_quantizer(stateful_rec.A_imag).values.numpy()
    variables_ak["A_imag"] = A_imag_ak.astype(np.int16)

    # Set intermediate internal_state_outputs_shift, which is a scale broadcasted to a vector
    internal_state_outputs_shift = stateful_rec.out_quantizer.shift.value.numpy()
    broadcast_and_set_variable(variables_ak, "internal_state_outputs_shift", np.full(
        A_real_ak.shape, internal_state_outputs_shift, dtype=np.int8))

    # Set out_proj dense kernel variable
    out_proj_ak = output_proj.weight_quantizer.qweights.value.fp.values.numpy()
    variables_ak["out_proj"] = out_proj_ak.astype(np.int8)

    if output_proj.use_bias:
        bias_quantizer = output_proj.bias_quantizer
        assert isinstance(bias_quantizer, qlayers.AlignedWeightQuantizer)
        # Set StatefuRecurrent layer bias variable and shift
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)
        variables_ak["bias"] = (bias >> bias_shift).astype(np.int8)
        broadcast_and_set_variable(variables_ak, "bias_shift", bias_shift)

    # Check if we have ReLU
    relu_layer = get_layer_by_type(block, qlayers.QuantizedReLU)
    # Set optional activation variables
    if relu_layer:
        set_relu_variables(ak_layer, relu_layer)

    # Get the layer block output_quantizer
    out_quantizer = get_block_out_quantizer(block)
    if out_quantizer:
        set_output_v2_variables(ak_layer, out_quantizer)


def _create_stateful_recurrent(block):
    """Parses a QuantizeML quantized stateful recurrent block and returns the
    params to create the corresponding Akida StatefulRecurrent layer.

    Args:
        block (list(:obj:`keras.Layer`)): list of quantized layers.

    Returns:
        :obj:`akida.StatefulRecurrent`: The created Akida layer.
    """
    input_proj = block[0]
    output_proj = block[3]

    # Compute reshaping factor from shapes
    upshape, downshape = 1, 1
    if input_proj.downshape:
        downshape = input_proj.weights[0].shape[0] // input_proj.input_shape[-1]
    if output_proj.upshape:
        upshape = output_proj.weights[0].shape[-1] // output_proj.upshape[-1]

    # In QuantizeML one bit is reserved automatically for the sign, but in Akida this is rather
    # checked during the clipping operations.
    block_params = dict(
        stateful_channels=input_proj.units,
        output_channels=output_proj.units,
        subsample=output_proj.subsample,
        downshape=downshape,
        upshape=upshape,
        activation=ActivationType.NoActivation,
        buffer_bits=input_proj.buffer_bitwidth + 1,
    )

    if relu_layer := get_layer_by_type(block, qlayers.QuantizedReLU):
        block_params.update(parse_relu_v2(relu_layer))

    parse_output_bits(block, block_params)
    parse_post_op_buffer_bits(block, block_params)
    return StatefulRecurrent(**block_params, name=output_proj.name)


def _convert_quantized_stateful_recurrent(model_ak, block, inbound_layers_ak):
    """Converts QuantizedLayerStatefulRecurrent layers block and its variables and adds
    it to the Akida's model.

    The expected sequence is:

    - QuantizedStatefulProjection,
    - QuantizedStatefulRecurrent,
    - QuantizedExtractToken,
    - QuantizedStatefulProjection,
    - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be added.
        block (list(:obj:`keras.Layer`)): list of quantizeml quantized layers.
        inbound_layers_ak (list(:obj:`akida.Layer`)): list of inbound Akida layers.
    """
    # Create and add layer to the akida model
    layer_ak = _create_stateful_recurrent(block)
    model_ak.add(layer_ak, inbound_layers_ak)

    # Set the akida layer converted variables
    _set_stateful_recurrent_variables(layer_ak, block)


class StatefulRecurrentBlockConverter(BlockConverter):
    """Main class that should be used to check if the stateful recurrent layer block is compatible
    to an Akida v2 conversion and provides a method to convert it in an equivalent Akida v2
    StatefulRecurrent layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._stateful_recurrent_additional_checks()

    def _stateful_recurrent_additional_checks(self):
        input_proj, output_proj = self._block[0], self._block[3]

        # Check input proj has no output quantizer, no bias, no activation
        assert input_proj.out_quantizer is None, "input_proj must not have output quantizer"
        assert not input_proj.use_bias, "input_proj must not use bias"

        # Check buffer bits match between input and output proj
        assert input_proj.buffer_bitwidth == output_proj.buffer_bitwidth, (
            "input_proj and output_proj buffer bits must match")

        # Shaping checks
        assert input_proj.subsample == 1, "input_proj must have subsample=1"
        assert input_proj.upshape is None, "input_proj must have upshape=None"
        assert output_proj.downshape is None, "output_proj must have downshape=None"

    def convert(self, model_ak, inbounds):
        _convert_quantized_stateful_recurrent(model_ak, self._block, inbounds)


# Register the valid stateful recurrent block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, StatefulRecurrentBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2025 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert an InputQuantizer into an Akida Quantizer layer.
"""

__all__ = ["InputQuantizerBlockConverter"]

import akida
from quantizeml.layers import InputQuantizer

from ..akida_versions import AkidaVersion
from .block_converter import BlockConverter, register_conversion_patterns
from .weights import broadcast_and_set_variable


_PATTERNS = [(InputQuantizer,)]


class InputQuantizerBlockConverter(BlockConverter):
    """Main class that should be used to convert an InputQuantizer in an
    equivalent Akida Quantizer layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def convert(self, model_ak, inbounds):
        input_quantizer = self._block[0]
        if (N := len(input_shape := tuple(input_quantizer.input_shape[1:]))) < 3:
            # Fill the missing dimensions with ones.
            input_shape = (1,) * (3 - N) + input_shape

        # Create the Akida Quantizer layer from the InputQuantizer layer.
        ak_quantizer = akida.Quantizer(
            input_shape=input_shape,
            output_bits=input_quantizer.bitwidth,
            output_signed=input_quantizer.signed,
            name=input_quantizer.name
        )

        # Add Quantizer into the Akida model.
        model_ak.add(ak_quantizer, inbounds)

        # Set layer variables.
        scales = 2.0 ** input_quantizer.frac_bits.value.numpy()
        broadcast_and_set_variable(ak_quantizer.variables, "scales", scales)
        if zero_points_recorder := getattr(input_quantizer, "zero_points", None):
            zero_points = zero_points_recorder.value.values.numpy().astype("uint8")
            broadcast_and_set_variable(ak_quantizer.variables, "zero_points", zero_points)


# Register the valid input conv block pattern for Akida v1 and v2
register_conversion_patterns(AkidaVersion.v1, _PATTERNS, InputQuantizerBlockConverter, True)
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, InputQuantizerBlockConverter, True)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["QuantizerOnnxConverter", "DequantizerOnnxConverter"]

import akida
import numpy as np

from .base_converter import OnnxConverter
from .weights import broadcast_and_set_variable
from .register import register_onnx_converter_target


@register_onnx_converter_target("InputQuantizer")
class QuantizerOnnxConverter(OnnxConverter):
    """Convert InputQuantizer node into an akida.Quantizer.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that the node is.
    """
    is_input_layer = True
    # Avoid to transpose channels, since akida.Quantizer will do it internally.
    channels_last = True

    def _additional_checks(self):
        super()._additional_checks()
        # Only channel swapping is allowed.
        N = len(self.perm)
        if self.perm not in [list(range(N)), [0, N - 1, *list(range(1, N - 1))]]:
            raise RuntimeError(f"Wrong permutation {self.perm}. Only channels may be permutted.")

    def _parse_akida_layer(self):
        # signed is determined by zero point type.
        signed = np.issubdtype(self.weights["zp"].dtype, np.signedinteger)
        return akida.Quantizer(input_shape=self.input_shape,
                               output_bits=8,
                               output_signed=signed,
                               channels_first=self.perm[1] == 1,
                               name=self.name)

    def _set_akida_variables(self, ak_layer):
        # Set scale as the inverse of the ONNX scale.
        broadcast_and_set_variable(ak_layer.variables, "scales", 1.0 / self.weights["scale"])
        # Cast zero point to uint8.
        zero_points = self.weights["zp"].astype(np.uint8)
        broadcast_and_set_variable(ak_layer.variables, "zero_points", zero_points)


@register_onnx_converter_target("Dequantizer")
class DequantizerOnnxConverter(OnnxConverter):
    """Convert Dequantizer node into an akida.Dequantizer.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that the node is.
    """

    def _additional_checks(self):
        super()._additional_checks()
        # Num inputs in onnx node may be weights + one input
        if len(self._node.input) != len(self.weights) + 1:
            raise RuntimeError(f"Multi-inbounds in {self.name} is not supported.")

    def _parse_akida_layer(self):
        return akida.Dequantizer(name=self.name)

    def _set_akida_variables(self, ak_layer):
        broadcast_and_set_variable(ak_layer.variables, "scales", self.weights["scale"])



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Parsing functions to generate an Akida model from a Keras model quantized with quantizeml api.
"""

from quantizeml.models import record_quantization_variables
from akida import Model, LayerType
from onnx import ModelProto

from .input_data import convert_input
from .layer_utils import get_inbound_layers
from .block_converters_generator import (generate_block_converters,
                                         _V1_INPUT_PATTERN_CONVERTERS,
                                         _V2_INPUT_PATTERN_CONVERTERS)
from .onnx_conversion import generate_onnx_model


def _raise_block_error(block_converter, exception):
    """ Raise error due to non convertible layers block"""
    block_layers_name = ""
    block_layers_type = ""
    for layer in block_converter._block:
        block_layers_name += f"'{layer.name}', "
        block_layers_type += f"{layer.__class__.__name__}, "
    raise type(exception)(f"Layers {block_layers_name[:-2]}: unsupported type "
                          f"[{block_layers_type[:-2]}]: {exception}") from exception


def generate_model(model):
    """Generates an Akida model.

    This function creates an Akida model by iterating through the layers of the
    quantized model. For each layer, the corresponding akida layer is created and
    added to the Akida model.

    Args:
        model (:obj:`keras.Model` or :obj:`onnx.ModelProto`): a model to convert.

    Returns:
        :obj:`akida.Model`: the generated Akida model.
    """
    if isinstance(model, ModelProto):
        return generate_onnx_model(model)
    return generate_qml_model(model)


def generate_qml_model(model):
    """Generates an Akida model based on a Keras quantizeml model.

    Args:
        model (:obj:`keras.Model`): a model to convert.

    Returns:
        :obj:`akida.Model`: the generated Akida model.
    """
    # Check if input model is convertible and extract its layers blocks
    blocks = generate_block_converters(model)
    # First store necessary variables for conversion
    record_quantization_variables(model)
    model_ak = Model()

    # Input blocks are handled with their associated BlockConverter
    INPUT_PATTERNS = (tuple(_V1_INPUT_PATTERN_CONVERTERS.values()) +
                      tuple(_V2_INPUT_PATTERN_CONVERTERS.values()))
    for converter in blocks:
        # The next check converts the layers block to its corresponding Akida layer
        try:
            inbounds = get_inbound_layers(model_ak, converter.root)
            if not isinstance(converter, INPUT_PATTERNS) and (
                len(model_ak.layers) == 0 or
                (len(inbounds) == 1 and inbounds[0].parameters.layer_type == LayerType.Quantizer)
            ):
                # Convert the keras InputLayer into an InputData layer
                inbounds = [convert_input(model_ak, model, inbounds)]
            converter.convert(model_ak, inbounds)
        except Exception as e:
            _raise_block_error(converter, e)
    return model_ak



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2025 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["BufferTempConvOnnxConverter"]

import akida

from .base_converter import OnnxConverter
from .register import register_onnx_converter_target
from .weights import set_weight_variables
from .scale_out import set_output_scale_variables
from .activation import set_relu_variables, parse_activation_type
from .conv_commons import _check_if_squared


@register_onnx_converter_target("QuantizedBufferTempConv")
class BufferTempConvOnnxConverter(OnnxConverter):
    """Convert QuantizedBufferTempConv type node into akida.BufferTempConv.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model containing the node.
    """
    @property
    def filters(self):
        return self.weights["W"].shape[0]

    def load_attributes(self, node):
        # Some attributes should infer from node.op_type
        n_op = node.op_type
        self.output_bits = 8 if "Scaled" in n_op else 32
        self.buffer_bits = 32
        self.post_op_buffer_bits = 32
        self.activation = parse_activation_type(n_op)
        return super().load_attributes(node)

    def _additional_checks(self):
        super()._additional_checks()
        kernel_shapes = self.weights["W"].shape[-2:]
        _check_if_squared(kernel_shapes, name=f"{self.name} kernels")

    def _parse_akida_layer(self):
        return akida.BufferTempConv(
            filters=self.filters,
            fifo_size=self.fifo_size,
            activation=self.activation,
            output_bits=self.output_bits,
            buffer_bits=self.buffer_bits,
            post_op_buffer_bits=self.post_op_buffer_bits,
            name=self.name)

    def _set_akida_variables(self, ak_layer):
        # Since the weights of the FifoOp conv are stored differently in quantizeml (F, C * T, H, W)
        # vs (H, W, T * C, F), transpose + reshape ops are performed to reorder the
        # (temporal/channels) dimensions of the fifo and to align the weights with akida.
        F, M, H, W = self.weights["W"].shape
        C = M // self.fifo_size
        kernel = self.weights["W"].reshape((F, C, self.fifo_size, H, W))
        kernel = kernel.transpose((3, 4, 2, 1, 0))
        kernel = kernel.reshape((H, W, M, F))
        ak_layer.variables["weights"] = kernel

        bias = self.weights.get("bias", None)
        set_weight_variables(ak_layer, kernel, bias)

        if self.activation == akida.ActivationType.ReLU:
            set_relu_variables(ak_layer, self.weights.get("max_value", None))

        # Scale out
        set_output_scale_variables(ak_layer, self.weights["Scale"], self.weights["Shift"])


@register_onnx_converter_target("QuantizedDepthwiseBufferTempConv")
class DepthwiseBufferTempConvOnnxConverter(BufferTempConvOnnxConverter):
    """Convert QuantizedDepthwiseBufferTempConv type node into akida.DepthwiseBufferTempConv.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model containing the node.
    """

    def _parse_akida_layer(self):
        return akida.DepthwiseBufferTempConv(
            fifo_size=self.fifo_size,
            activation=self.activation,
            output_bits=self.output_bits,
            buffer_bits=self.buffer_bits,
            post_op_buffer_bits=self.post_op_buffer_bits,
            name=self.name)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["generate_onnx_model"]
import akida
import warnings
from onnx.checker import check_model

from .register import map_node_to_converter


def generate_onnx_model(model):
    """Generates an Akida model based on an ONNX quantizeml model.

    Args:
        model (obj:`onnx.ModelProto`): a ONNX model to convert.

    Returns:
        akida.Model: the generated Akida model.
    """
    # Model must be compatible with ONNX.
    check_model(model, full_check=True)

    # Checks over model.
    if len(model.graph.input) != 1 or len(model.graph.output) != 1:
        raise ValueError("Cannot convert: model must have exactly one input and one output.")

    # Main conversion loop.
    akida_model = akida.Model()
    for idx, node in enumerate(model.graph.node):
        try:
            converter = map_node_to_converter(node, model)
            converter.convert(akida_model)
            if node.op_type == "Dequantizer":
                # End conversion at the first dequantizer found.
                idx += 1
                break
        except Exception as e:
            raise type(e)(f"Cannot convert {node.name}: {e}") from e

    # Check model is not empty (at least one node to convert is required).
    if akida.graph_utils.first_mappable_layer(akida_model.layers) is None:
        raise ValueError("Model is empty or does not have any node to convert.")

    # Prints a warning message with a summary of the non convertible nodes
    skip_nodes = model.graph.node[idx:]
    if len(skip_nodes) > 0:
        stop_layer_msg = " at node " + model.graph.node[idx].name
        skip_layers_summary = "___________________________________________________\n"
        skip_layers_summary += "Node (type)\n"
        skip_layers_summary += "===================================================\n"
        for node in skip_nodes:
            skip_layers_summary += node.name + " (" + node.op_type + ")\n"
        skip_layers_summary += "===================================================\n"
        warnings.warn(f"Conversion stops {stop_layer_msg} because of a dequantizer. "
                      f"The end of the graph is ignored:\n {skip_layers_summary}. \n This can be "
                      "expected for model heads (e.g. softmax for classification) but could also "
                      "mean that processing layers were not quantized.")
    return akida_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["get_akida_padding", "compute_conv_transpose_same_pads", "compute_conv_pads"]

import math
import akida
from quantizeml.onnx_support.layers.compute_shapes import compute_conv_output


def _check_same_upper_padding_constraints(converter):
    kernel_shape = converter.weights["W"].shape[-2:]
    allowed_kernels = {(3, 3), (5, 5), (7, 7)} if converter.is_input_layer else {(3, 3)}
    allowed_stride = [2, 2]

    if converter.strides != allowed_stride:
        raise ValueError(
            f"'SameUpper' padding mode is only supported for "
            f"strides {allowed_stride} in {converter.name}. Got strides {converter.strides}."
        )
    if kernel_shape not in allowed_kernels:
        raise ValueError(
            f"'SameUpper' padding mode is only supported for "
            f"kernel sizes {allowed_kernels} for {converter.name}. "
            f"Got kernel size {kernel_shape}."
        )

    if getattr(converter, "pool_type", None) == akida.PoolType.Max:
        raise ValueError(
            f"MaxPool is not supported with 'SameUpper' padding mode in {converter.name}.")


def compute_conv_pads(input_shape, kernel_shape, strides, padding_mode="same_lower"):
    """Compute pads values.

    Args:
        input_shape (tuple of ints): the input shape.
        kernel_shape (tuple of ints): the convolutional kernel shape.
        strides (tuple of ints): the convolutional strides.
        padding_mode (str): the padding mode, one of "valid", "same_lower", "same_upper",
            "same_symmetric". Defaults to "same_lower".

    Returns:
        tuple of ints: the pads to apply.
    """
    x, y = input_shape
    filter_x, filter_y = kernel_shape
    stride_x, stride_y = strides

    if x % stride_x == 0:
        pad_along_x = max(filter_x - stride_x, 0)
    else:
        pad_along_x = max(filter_x - (x % stride_x), 0)
    if y % stride_y == 0:
        pad_along_y = max(filter_y - stride_y, 0)
    else:
        pad_along_y = max(filter_y - (y % stride_y), 0)
    pad_y1 = pad_along_y // 2
    pad_y2 = pad_along_y - pad_y1
    pad_x1 = pad_along_x // 2
    pad_x2 = pad_along_x - pad_x1

    match padding_mode:
        case "valid":
            return [0, 0, 0, 0]
        case "same_upper":
            return [pad_x2, pad_y2, pad_x1, pad_y1]
        case "same_lower":
            return [pad_x1, pad_y1, pad_x2, pad_y2]
        case "same_symmetric":
            # As for pytorch
            return [pad_x2, pad_y2, pad_x2, pad_y2]
        case _:
            raise ValueError(f"Unknown padding_mode '{padding_mode}'")


def compute_conv_transpose_same_pads(kernel_shape, strides):
    """Compute pads values for conv transpose
    See https://onnx.ai/onnx/operators/onnx__ConvTranspose.html#summary
    The Akida "SAME" padding corresponds to "SAME_LOWER" in ONNX.

    Args:
        kernel_shape (list or tuple of ints): the convolutional kernel shape.
        strides (list or tuple of ints): the convolutional strides.

    Returns:
        list of ints: the pads to apply.
    """
    filter_x, filter_y = kernel_shape
    stride_x, stride_y = strides

    total_padding_x = max(filter_x - stride_x, 0)
    total_padding_y = max(filter_y - stride_y, 0)

    pad_x2 = math.ceil(total_padding_x / 2)
    pad_x1 = total_padding_x - pad_x2

    pad_y2 = math.ceil(total_padding_y / 2)
    pad_y1 = total_padding_y - pad_y2

    return [pad_x1, pad_y1, pad_x2, pad_y2]


def get_akida_padding(converter):
    """Returns the akida padding

    Args:
        converter (OnnxConverter): the converter to extract the padding.

    Returns:
        akida.Padding: the padding
    """
    # Check pads are compatible with akida.Padding.Same
    kernel_shapes = converter.weights["W"].shape[-2:]
    # Compute expected pads
    exp_pads = []
    for pad_mode in ["valid", "same_lower", "same_upper", "same_symmetric"]:
        exp_pads.append(compute_conv_pads(converter.input_shape[:2], kernel_shapes,
                                          converter.strides, pad_mode))

    # Pads in FC dimensions have to be zero
    fc_pads = converter.pads[:2] + converter.pads[4:6]
    if any(fc_pads):
        raise ValueError(f"Unrecognized {converter.pads} pads in {converter.name}.")
    # Compare if convolutional padding produces same behavior than Akida
    convert_pads = converter.pads[2:4] + converter.pads[6:]

    # Check for Same padding
    if convert_pads == exp_pads[1]:
        ak_padding = akida.Padding.Same

    # Check for SameUpper or Symmmetric padding
    elif convert_pads in exp_pads[2:]:
        _check_same_upper_padding_constraints(converter)
        ak_padding = akida.Padding.SameUpper

    # Check for Valid padding
    elif convert_pads == exp_pads[0]:
        # Force ak_padding to Valid, because no padding is supported as Valid in akida HW
        ak_padding = akida.Padding.Valid

    else:
        raise ValueError(
            f"Expect pads to be one of {exp_pads} (found {convert_pads}) in {converter.name}.")

    # Compare if max pool padding produces same behavior than Akida
    if converter.pool_type == akida.PoolType.Max:
        # Calculate convolutional output shape
        out_shape = compute_conv_output(converter.input_shape[:2],
                                        kernel_shapes,
                                        converter.strides,
                                        convert_pads)
        if ak_padding == akida.Padding.Same:
            exp_pool_pads = compute_conv_pads(out_shape,
                                              converter.pool_size,
                                              converter.pool_strides)
            if converter.pool_pads != exp_pool_pads and all(x == 0 for x in exp_pads):
                # Mismatch when padding Same. Try once again with a valid one,
                # given in this condition both padding are exchangable.
                ak_padding = akida.Padding.Valid
            else:
                exp_pads = exp_pool_pads
        if converter.pool_pads != exp_pads:
            raise ValueError(f"Expect pads {exp_pads} (found {converter.pool_pads}) "
                             f"in {converter.name} maxpool.")

    return ak_padding
